@@ -1,4 +1,4 @@
-# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
+# Copyright (C) 2012-2021 Ben Kurtovic <ben.kurtovic@gmail.com>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -24,11 +24,17 @@ import re
 
 from . import contexts, tokens
 from .errors import ParserError
-from ..definitions import (get_html_tag, is_parsable, is_single,
-                           is_single_only, is_scheme)
+from ..definitions import (
+    get_html_tag,
+    is_parsable,
+    is_single,
+    is_single_only,
+    is_scheme,
+)
 
 __all__ = ["Tokenizer"]
 
+
 class BadRoute(Exception):
     """Raised internally when the current tokenization route is invalid."""
 
@@ -39,14 +45,15 @@ class BadRoute(Exception):
 
 class _TagOpenData:
     """Stores data about an HTML open tag, like ``<ref name="foo">``."""
-    CX_NAME =        1 << 0
-    CX_ATTR_READY =  1 << 1
-    CX_ATTR_NAME =   1 << 2
-    CX_ATTR_VALUE =  1 << 3
-    CX_QUOTED =      1 << 4
-    CX_NOTE_SPACE =  1 << 5
+
+    CX_NAME = 1 << 0
+    CX_ATTR_READY = 1 << 1
+    CX_ATTR_NAME = 1 << 2
+    CX_ATTR_VALUE = 1 << 3
+    CX_QUOTED = 1 << 4
+    CX_NOTE_SPACE = 1 << 5
     CX_NOTE_EQUALS = 1 << 6
-    CX_NOTE_QUOTE =  1 << 7
+    CX_NOTE_QUOTE = 1 << 7
 
     def __init__(self):
         self.context = self.CX_NAME
@@ -57,11 +64,34 @@ class _TagOpenData:
 
 class Tokenizer:
     """Creates a list of tokens from a string of wikicode."""
+
     USES_C = False
     START = object()
     END = object()
-    MARKERS = ["{", "}", "[", "]", "<", ">", "|", "=", "&", "'", "#", "*", ";",
-               ":", "/", "-", "!", "\n", START, END]
+    MARKERS = [
+        "{",
+        "}",
+        "[",
+        "]",
+        "<",
+        ">",
+        "|",
+        "=",
+        "&",
+        "'",
+        '"',
+        "#",
+        "*",
+        ";",
+        ":",
+        "/",
+        "-",
+        "!",
+        "\n",
+        START,
+        END,
+    ]
+    URISCHEME = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+.-"
     MAX_DEPTH = 40
     regex = re.compile(r"([{}\[\]<>|=&'#*;:/\\\"\-!\n])", flags=re.IGNORECASE)
     tag_splitter = re.compile(r"([\s\"\'\\]+)")
@@ -323,7 +353,7 @@ class Tokenizer:
         self._head += 2
         try:
             # If the wikilink looks like an external link, parse it as such:
-            link, _extra, _delta = self._really_parse_external_link(True)
+            link, _extra = self._really_parse_external_link(True)
         except BadRoute:
             self._head = reset + 1
             try:
@@ -366,8 +396,7 @@ class Tokenizer:
             self._emit_text("//")
             self._head += 2
         else:
-            valid = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+.-"
-            all_valid = lambda: all(char in valid for char in self._read())
+            all_valid = lambda: all(char in self.URISCHEME for char in self._read())
             scheme = ""
             while self._read() is not self.END and all_valid():
                 scheme += self._read()
@@ -386,17 +415,16 @@ class Tokenizer:
 
     def _parse_free_uri_scheme(self):
         """Parse the URI scheme of a free (no brackets) external link."""
-        valid = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+.-"
         scheme = []
         try:
             # We have to backtrack through the textbuffer looking for our
             # scheme since it was just parsed as text:
             for chunk in reversed(self._textbuffer):
                 for char in reversed(chunk):
-                    # stop at the first non-word character
+                    # Stop at the first non-word character
                     if re.fullmatch(r"\W", char):
                         raise StopIteration()
-                    if char not in valid:
+                    if char not in self.URISCHEME:
                         raise BadRoute()
                     scheme.append(char)
         except StopIteration:
@@ -434,23 +462,26 @@ class Tokenizer:
         self._emit_text(this)
         return punct, tail
 
-    def _is_free_link_end(self, this, nxt):
-        """Return whether the current head is the end of a free link."""
+    def _is_uri_end(self, this, nxt):
+        """Return whether the current head is the end of a URI."""
         # Built from _parse()'s end sentinels:
         after, ctx = self._read(2), self._context
-        equal_sign_contexts = contexts.TEMPLATE_PARAM_KEY | contexts.HEADING
-        return (this in (self.END, "\n", "[", "]", "<", ">", "\"") or
-                this == nxt == "'" or
-                (this == "|" and ctx & contexts.TEMPLATE) or
-                (this == "=" and ctx & equal_sign_contexts) or
-                (this == nxt == "}" and ctx & contexts.TEMPLATE) or
-                (this == nxt == after == "}" and ctx & contexts.ARGUMENT))
+        return (
+            this in (self.END, "\n", "[", "]", "<", ">", '"')
+            or " " in this
+            or this == nxt == "'"
+            or (this == "|" and ctx & contexts.TEMPLATE)
+            or (this == "=" and ctx & (contexts.TEMPLATE_PARAM_KEY | contexts.HEADING))
+            or (this == nxt == "}" and ctx & contexts.TEMPLATE)
+            or (this == nxt == after == "}" and ctx & contexts.ARGUMENT)
+        )
 
     def _really_parse_external_link(self, brackets):
         """Really parse an external link."""
         if brackets:
             self._parse_bracketed_uri_scheme()
             invalid = ("\n", " ", "]")
+            punct = ()
         else:
             self._parse_free_uri_scheme()
             invalid = ("\n", " ", "[", "]")
@@ -465,53 +496,47 @@ class Tokenizer:
                     self._emit_text(tail)
                     tail = ""
                 self._parse_entity()
-            elif (this == "<" and nxt == "!" and self._read(2) ==
-                  self._read(3) == "-"):
+            elif this == "<" and nxt == "!" and self._read(2) == self._read(3) == "-":
                 if tail:
                     self._emit_text(tail)
                     tail = ""
                 self._parse_comment()
-            elif not brackets and self._is_free_link_end(this, nxt):
-                return self._pop(), tail, -1
-            elif this is self.END or this == "\n":
-                self._fail_route()
             elif this == nxt == "{" and self._can_recurse():
                 if tail:
                     self._emit_text(tail)
                     tail = ""
                 self._parse_template_or_argument()
-            elif this == "]":
-                return self._pop(), tail, 0
-            elif this == "'" and nxt == "'":
-                separator = tokens.ExternalLinkSeparator()
-                separator.suppress_space = True
-                self._emit(separator)
-                self._context ^= contexts.EXT_LINK_URI
-                self._context |= contexts.EXT_LINK_TITLE
-                return self._parse(push=False), None, 0
-            elif any(ch in this for ch in (" ", "\n", "[", "]", "<", ">",
-                                           "\"")):
-                before, after = re.split(r"[ \n[\]<>\"]", this, maxsplit=1)
-                delimiter = this[len(before)]
-                if brackets:
-                    self._emit_text(before)
-                    separator = tokens.ExternalLinkSeparator()
-                    if delimiter != " ":
+            elif brackets:
+                if this is self.END or this == "\n":
+                    self._fail_route()
+                if this == "]":
+                    return self._pop(), None
+                if self._is_uri_end(this, nxt):
+                    if " " in this:
+                        before, after = this.split(" ", 1)
+                        self._emit_text(before)
+                        self._emit(tokens.ExternalLinkSeparator())
+                        if after:
+                            self._emit_text(after)
+                        self._head += 1
+                    else:
+                        separator = tokens.ExternalLinkSeparator()
                         separator.suppress_space = True
-                    self._emit(separator)
-                    if after:
-                        self._emit_text(after)
+                        self._emit(separator)
                     self._context ^= contexts.EXT_LINK_URI
                     self._context |= contexts.EXT_LINK_TITLE
-                    if delimiter == " ":
-                        self._head += 1
-                    return self._parse(push=False), None, 0
-                punct, tail = self._handle_free_link_text(punct, tail, before)
-                return self._pop(), tail + " " + after, 0
-            elif not brackets:
-                punct, tail = self._handle_free_link_text(punct, tail, this)
-            else:
+                    return self._parse(push=False), None
                 self._emit_text(this)
+            else:
+                if self._is_uri_end(this, nxt):
+                    if this is not self.END and " " in this:
+                        before, after = this.split(" ", 1)
+                        punct, tail = self._handle_free_link_text(punct, tail, before)
+                        tail += " " + after
+                    else:
+                        self._head -= 1
+                    return self._pop(), tail
+                punct, tail = self._handle_free_link_text(punct, tail, this)
             self._head += 1
 
     def _remove_uri_scheme_from_textbuffer(self, scheme):
@@ -536,7 +561,7 @@ class Tokenizer:
         reset = self._head
         self._head += 1
         try:
-            link, extra, delta = self._really_parse_external_link(brackets)
+            link, extra = self._really_parse_external_link(brackets)
         except BadRoute:
             self._head = reset
             if not brackets and self._context & contexts.DL_TERM:
@@ -550,7 +575,6 @@ class Tokenizer:
             self._emit(tokens.ExternalLinkOpen(brackets=brackets))
             self._emit_all(link)
             self._emit(tokens.ExternalLinkClose())
-            self._head += delta
             if extra:
                 self._emit_text(extra)
 
@@ -688,9 +712,13 @@ class Tokenizer:
             self._emit_first(tokens.TagAttrQuote(char=data.quoter))
             self._emit_all(self._pop())
         buf = data.padding_buffer
-        self._emit_first(tokens.TagAttrStart(
-            pad_first=buf["first"], pad_before_eq=buf["before_eq"],
-            pad_after_eq=buf["after_eq"]))
+        self._emit_first(
+            tokens.TagAttrStart(
+                pad_first=buf["first"],
+                pad_before_eq=buf["before_eq"],
+                pad_after_eq=buf["after_eq"],
+            )
+        )
         self._emit_all(self._pop())
         for key in data.padding_buffer:
             data.padding_buffer[key] = ""
@@ -698,7 +726,9 @@ class Tokenizer:
     def _handle_tag_space(self, data, text):
         """Handle whitespace (*text*) inside of an HTML open tag."""
         ctx = data.context
-        end_of_value = ctx & data.CX_ATTR_VALUE and not ctx & (data.CX_QUOTED | data.CX_NOTE_QUOTE)
+        end_of_value = ctx & data.CX_ATTR_VALUE and not ctx & (
+            data.CX_QUOTED | data.CX_NOTE_QUOTE
+        )
         if end_of_value or (ctx & data.CX_QUOTED and ctx & data.CX_NOTE_SPACE):
             self._push_tag_buffer(data)
             data.context = data.CX_ATTR_READY
@@ -799,8 +829,10 @@ class Tokenizer:
         """Handle the ending of a closing tag (``</foo>``)."""
         strip = lambda tok: tok.text.rstrip().lower()
         closing = self._pop()
-        if len(closing) != 1 or (not isinstance(closing[0], tokens.Text) or
-                                 strip(closing[0]) != strip(self._stack[1])):
+        if len(closing) != 1 or (
+            not isinstance(closing[0], tokens.Text)
+            or strip(closing[0]) != strip(self._stack[1])
+        ):
             self._fail_route()
         self._emit_all(closing)
         self._emit(tokens.TagCloseClose())
@@ -815,8 +847,9 @@ class Tokenizer:
                 self._fail_route()
             elif this == "<" and nxt == "/":
                 self._head += 3
-                if self._read() != ">" or (strip(self._read(-1)) !=
-                                           strip(self._stack[1].text)):
+                if self._read() != ">" or (
+                    strip(self._read(-1)) != strip(self._stack[1].text)
+                ):
                     self._head -= 1
                     self._emit_text("</")
                     continue
@@ -854,8 +887,8 @@ class Tokenizer:
                 depth -= 1
                 if depth == 0:  # pragma: no cover (untestable/exceptional)
                     raise ParserError(
-                        "_handle_single_tag_end() got an unexpected "
-                        "TagCloseSelfclose")
+                        "_handle_single_tag_end() got an unexpected TagCloseSelfclose"
+                    )
         else:  # pragma: no cover (untestable/exceptional case)
             raise ParserError("_handle_single_tag_end() missed a TagCloseOpen")
         padding = stack[index].padding
@@ -869,8 +902,10 @@ class Tokenizer:
         self._emit(tokens.TagOpenOpen())
         while True:
             this, nxt = self._read(), self._read(1)
-            can_exit = (not data.context & (data.CX_QUOTED | data.CX_NAME) or
-                        data.context & data.CX_NOTE_SPACE)
+            can_exit = (
+                not data.context & (data.CX_QUOTED | data.CX_NAME)
+                or data.context & data.CX_NOTE_SPACE
+            )
             if this is self.END:
                 if self._context & contexts.TAG_ATTR:
                     if data.context & data.CX_QUOTED:
@@ -1086,16 +1121,25 @@ class Tokenizer:
         else:
             self._emit_text("\n")
 
-    def _emit_table_tag(self, open_open_markup, tag, style, padding,
-                        close_open_markup, contents, open_close_markup):
+    def _emit_table_tag(
+        self,
+        open_open_markup,
+        tag,
+        style,
+        padding,
+        close_open_markup,
+        contents,
+        open_close_markup,
+    ):
         """Emit a table tag."""
         self._emit(tokens.TagOpenOpen(wiki_markup=open_open_markup))
         self._emit_text(tag)
         if style:
             self._emit_all(style)
         if close_open_markup:
-            self._emit(tokens.TagCloseOpen(wiki_markup=close_open_markup,
-                                           padding=padding))
+            self._emit(
+                tokens.TagCloseOpen(wiki_markup=close_open_markup, padding=padding)
+            )
         else:
             self._emit(tokens.TagCloseOpen(padding=padding))
         if contents:
@@ -1110,8 +1154,9 @@ class Tokenizer:
         data.context = _TagOpenData.CX_ATTR_READY
         while True:
             this = self._read()
-            can_exit = (not data.context & data.CX_QUOTED or
-                        data.context & data.CX_NOTE_SPACE)
+            can_exit = (
+                not data.context & data.CX_QUOTED or data.context & data.CX_NOTE_SPACE
+            )
             if this == end_token and can_exit:
                 if data.context & (data.CX_ATTR_NAME | data.CX_ATTR_VALUE):
                     self._push_tag_buffer(data)
@@ -1194,30 +1239,34 @@ class Tokenizer:
             self._head -= 1
             return
 
-        cell = self._parse(contexts.TABLE_OPEN | contexts.TABLE_CELL_OPEN |
-                           line_context | contexts.TABLE_CELL_STYLE)
+        cell = self._parse(
+            contexts.TABLE_OPEN
+            | contexts.TABLE_CELL_OPEN
+            | line_context
+            | contexts.TABLE_CELL_STYLE
+        )
         cell_context = self._context
         self._context = old_context
         reset_for_style = cell_context & contexts.TABLE_CELL_STYLE
         if reset_for_style:
             self._head = reset
-            self._push(contexts.TABLE_OPEN | contexts.TABLE_CELL_OPEN |
-                       line_context)
+            self._push(contexts.TABLE_OPEN | contexts.TABLE_CELL_OPEN | line_context)
             padding = self._handle_table_style("|")
             style = self._pop()
             # Don't parse the style separator:
             self._head += 1
-            cell = self._parse(contexts.TABLE_OPEN | contexts.TABLE_CELL_OPEN |
-                               line_context)
+            cell = self._parse(
+                contexts.TABLE_OPEN | contexts.TABLE_CELL_OPEN | line_context
+            )
             cell_context = self._context
             self._context = old_context
 
         close_open_markup = "|" if reset_for_style else None
-        self._emit_table_tag(markup, tag, style, padding, close_open_markup,
-                             cell, "")
+        self._emit_table_tag(markup, tag, style, padding, close_open_markup, cell, "")
         # Keep header/cell line contexts:
-        self._context |= cell_context & (contexts.TABLE_TH_LINE |
-                                         contexts.TABLE_TD_LINE)
+        self._context |= cell_context & (
+            contexts.TABLE_TH_LINE | contexts.TABLE_TD_LINE
+        )
         # Offset displacement done by parse():
         self._head -= 1
 
@@ -1340,7 +1389,11 @@ class Tokenizer:
             elif this == "|" and self._context & contexts.TEMPLATE:
                 self._handle_template_param()
             elif this == "=" and self._context & contexts.TEMPLATE_PARAM_KEY:
-                if not self._global & contexts.GL_HEADING and self._read(-1) in ("\n", self.START) and nxt == "=":
+                if (
+                    not self._global & contexts.GL_HEADING
+                    and self._read(-1) in ("\n", self.START)
+                    and nxt == "="
+                ):
                     self._parse_heading()
                 else:
                     self._handle_template_param_value()
@@ -1369,7 +1422,11 @@ class Tokenizer:
                 self._parse_external_link(False)
             elif this == "]" and self._context & contexts.EXT_LINK_TITLE:
                 return self._pop()
-            elif this == "=" and not self._global & contexts.GL_HEADING and not self._context & contexts.TEMPLATE:
+            elif (
+                this == "="
+                and not self._global & contexts.GL_HEADING
+                and not self._context & contexts.TEMPLATE
+            ):
                 if self._read(-1) in ("\n", self.START):
                     self._parse_heading()
                 else:
@@ -1404,7 +1461,8 @@ class Tokenizer:
             elif self._read(-1) in ("\n", self.START) and this in ("#", "*", ";", ":"):
                 self._handle_list()
             elif self._read(-1) in ("\n", self.START) and (
-                    this == nxt == self._read(2) == self._read(3) == "-"):
+                this == nxt == self._read(2) == self._read(3) == "-"
+            ):
                 self._handle_hr()
             elif this in ("\n", ":") and self._context & contexts.DL_TERM:
                 self._handle_dl_term()
@@ -1412,9 +1470,17 @@ class Tokenizer:
                     # Kill potential table contexts
                     self._context &= ~contexts.TABLE_CELL_LINE_CONTEXTS
             # Start of table parsing
-            elif this == "{" and nxt == "|" and (
-                    self._read(-1) in ("\n", self.START) or
-                    (self._read(-2) in ("\n", self.START) and self._read(-1).isspace())):
+            elif (
+                this == "{"
+                and nxt == "|"
+                and (
+                    self._read(-1) in ("\n", self.START)
+                    or (
+                        self._read(-2) in ("\n", self.START)
+                        and self._read(-1).isspace()
+                    )
+                )
+            ):
                 if self._can_recurse():
                     self._parse_table()
                 else:
@@ -1438,8 +1504,9 @@ class Tokenizer:
                 elif this == "\n" and self._context & contexts.TABLE_CELL_LINE_CONTEXTS:
                     self._context &= ~contexts.TABLE_CELL_LINE_CONTEXTS
                     self._emit_text(this)
-                elif (self._read(-1) in ("\n", self.START) or
-                      (self._read(-2) in ("\n", self.START) and self._read(-1).isspace())):
+                elif self._read(-1) in ("\n", self.START) or (
+                    self._read(-2) in ("\n", self.START) and self._read(-1).isspace()
+                ):
                     if this == "|" and nxt == "}":
                         if self._context & contexts.TABLE_CELL_OPEN:
                             return self._handle_table_cell_end()
